;;;###autoload
(package-provide 'xemacs-base
		 :version 1.87
          :author-version "No-Upstream-Ver"
		 :type 'regular)

;;;###autoload
(when (fboundp 'package-suppress)
      (package-suppress 'xemacs-base "regexp-opt" '(emacs-version>= 21 5 11)) (package-suppress 'xemacs-base "easy-mmode" '(emacs-version>= 21 5 11))
)
